from learner import Learner
import torch
import random

# train the gamer

W = 10
H = 10

class HalfBattleship:
    def __init__(self):
        self.W = W
        self.H = H
        self.board = [None for i in range(W*H)]
        self.hits = [None for i in range(W*H)]
        self.pieces = []
        for l in [2,3,4,5,6]:
            while True:
                h = random.randint(0, 1) == 1
                if h:
                    x = random.randrange(0, self.W - l + 1)
                    y = random.randrange(0, self.H)
                    row = [(x+dx,y)
                           for dx in range(0, l)]
                else:
                    x = random.randrange(0, self.W)
                    y = random.randrange(0, self.H - l + 1)
                    row = [(x,y+dy)
                           for dy in range(0, l)]
                if len([v for v in filter(lambda o: self.get(o[0], o[1])[0] is not None, row)]) == 0:
                    for x, y in row:
                        self.board[x + y*self.W] = True
                    self.pieces.append((x,y,h))
                    break

    def get(self, x, y):
        return (self.board[x + y*self.W], self.hits[x + y*self.W])

    def state(self):
        return [item for it in [
            [
                1 if x is not None and x else 0,
                1 if x is not None and not x else 0
            ]
            for x in self.hits
        ] for item in it]

    def actions(self):
        return [(x,y) if self.get(x,y)[1] is None else None
                for y in range(self.H)
                for x in range(self.W)]

    def get_str(self, x, y):
        piece, hit = self.get(x, y)
        if hit is not None:
            return '#' if hit else '.'
        return 'o' if piece is not None else ' '

    def print(self, other):
        if self.W != other.W or self.H != other.H:
            raise ValueError('dimensions unequal')

        print('/'+''.join(['-' for i in range(0,self.W+self.W-1)])+'\\   /'+
                  ''.join(['-' for i in range(0,other.W+other.W-1)])+'\\')
        for y in range(self.H):
            print("|" + (' '.join([self.get_str(x,y)
                                    for x in range(self.W)])) + "|   |" +
                          (' '.join([other.get_str(x,y)
                                    for x in range(other.W)]))+'|')
        print('\\'+''.join(['-' for i in range(0,self.W+self.W-1)])+'/   \\'+
                  ''.join(['-' for i in range(0,other.W+other.W-1)])+'/')

    def play(self, x, y):
        piece, hit = self.get(x, y)

        if hit is not None:
            return None

        hit = True if piece else False
        self.hits[x + y*self.W] = hit

        if len([v for v in filter(lambda v: v[0] is not None and v[1] is None,
                      [self.get(x,y)
                       for y in range(self.H)
                       for x in range(self.W)])]) == 0:
            return (True, True)
        return (False, hit)

if __name__ == '__main__':

    state = None

    try:
        state = torch.load("model-bship.pt", weights_only = True)
    except:
        pass

    players = [1,2]
    games = [HalfBattleship() for pl in players]
    states = [None for pl in players]
    acts = [None for pl in players]
    next_states = [None for pl in players]

    agent = Learner(len(games[0].actions()), len(games[0].state()), state = state, EPS_DECAY=10000)

    while True:
        for i, pl in enumerate(players):
            print()
            game = games[i]
            states[i] = game.state()
            next_states[i] = None
            state = states[i]

            act = None
            actions = game.actions()
            while act is None:
                act = agent.decide(state)
                if actions[act] is None:
                    print('badact', act)
                    agent.learn(state, act, state, -1)
                    act = None

            acts[i] = act

            term, winner = game.play(actions[act][0], actions[act][1])
            print("state:",(term,winner))
            next_states[i] = game.state()
            print("learn", 1 if winner else -0.1)
            agent.learn(states[i], acts[i], next_states[i], 1 if winner else 0)

            other_reward = 0
            if term and winner:
                print()
                print("@@@@")
                print("@@@@")
                print("@@@@ WINNER :"+str(pl)+" @@@@")
                print("@@@@")
                print("@@@@")
                print()

            coverage = len([v for v in filter(lambda v: v is not None, game.hits)]) / len(game.hits)
            print(str(round(coverage*100))+"% covered")
            games[0].print(games[1])
            torch.save(agent.get_model_state(), "model-bship.pt")

            if term or coverage > 0.6:
                games = [HalfBattleship() for pl in players]
                states = [None for pl in players]
                acts = [None for pl in players]
                next_states = [None for pl in players]
                break
